// TableCalc_Format.c -*- c++ -*- Time-stamp: <2008/03/23>
// (C) 2005-2008 HAL <http://hal9.blog3.fc2.com> HAL.Kinosita@gmail.com

// double -> string ւ̕ϊBPCalc/ConvDblToStr.c 

// GPL ɊÂC̃\[XR[h̎gpCρCĔzz͎RłB

#include <PalmOS.h>
//#include <StdIOPalm.h>

#include "MathLib.h"
#include "TableCalc.h"
#include "TableCalc_rsc.h"

#define FIX_FORMAT     0x01
#define EXP_FORMAT     0x02
#define ADD_SEP_INTP   0x04
#define ADD_SEP_FRACP  0x08

//=======================================================================================================================================
//=======================================================================================================================================
static char ThousSepIntPart  = '\'';
static char ThousSepFracPart = '\'';

// prototypes for internal functions. 
static int  ConvDblToStr_f(double, char*, int);
static int  ConvDblToStr_e(double, char*, int);
static int  ConvDblToStr_i(double, char*);
static void AddThousandSeparator(char* s);


// function definitions. 
void SetThausandSeparator(char s1, char s2){
  ThousSepIntPart  = s1;
  ThousSepFracPart = s2;
}

int ConvertDblToStr(double x, char* s, int format, int digit){
  int ret;
  char s1, s2;

  if (isnan(x)) {
    StrCopy(s, "[NaN]");
    return 1;
  }
  if (isinf(x) == +1) {
    StrCopy(s, "[+INF]");
    return 1;
  }
  if (isinf(x) == -1) {
    StrCopy(s, "[-INF]");
    return 1;
  }

  s1 = ThousSepIntPart;
  s2 = ThousSepFracPart;
  if (!(format & ADD_SEP_INTP))  ThousSepIntPart  = '\0';
  if (!(format & ADD_SEP_FRACP)) ThousSepFracPart = '\0';

  if (format & FIX_FORMAT) {
    ret = ConvDblToStr_f(x, s, digit);
    if (ret == 0) AddThousandSeparator(s);
  } else if (format & EXP_FORMAT) {
    ret = ConvDblToStr_e(x, s, digit);
  } else {
    ret = 1;
  }

  ThousSepIntPart  = s1;
  ThousSepFracPart = s2;
  return ret;
}
   

int ConvDblToStr_f(double x, char* s, int d)
{
  double integer, fraction;
  char* ptr = s;
  int iii;

  if (x < 0) {
    *(ptr++) = '-';
    x = fabs(x);
  }

  // divide into int. & frac. 
  x += pow(10, -d) * 0.5;
  fraction = modf(x, &integer);

  // integer part 
  ConvDblToStr_i(integer, ptr);
  ptr = s + StrLen(s);
  *(ptr++) = '.';

  // fraction part 
  fraction = fraction * pow(10, d);
  iii = d - (fraction > 10 ? (int)log10(fraction) + 1 : 1);
  while (iii-- > 0) *(ptr++) = '0';

  ConvDblToStr_i(fraction, ptr);
  ptr = s + StrLen(s);
  *ptr = '\0';

  // suppress trailing zeros. 
  while (*(ptr - 1) == '0') *(--ptr) = '\0';
  if (*(ptr - 1) == '.') *(--ptr) = '\0';

  return 0;
}


int ConvDblToStr_e(double x, char* s, int d)
{
  double mantissa, exponent;
  char* ptr = s;

  if (fabs(x) < 1e-308) {
    exponent = 0;
    mantissa = 0;
  } else if (x > 0) {
    exponent = floor(log10(x));
    mantissa = x / pow(10, exponent);
  } else {
    exponent = floor(log10(-x));
    mantissa = x / pow(10, exponent);
  }

  ConvDblToStr_f(mantissa, s, d);
  AddThousandSeparator(s);
  ptr += StrLen(s);
  *(ptr++) = 'e';

  if (exponent >= 0) *(ptr++) = '+';

  ConvDblToStr_f(exponent, ptr, 3);
  return 0;
}


int ConvDblToStr_i(double x, char* s)
{
  int lg10, i;
  unsigned long y;
  char* ptr = s;

  lg10 = (x > 9.) ? (int)log10(x) : 0;
  if (lg10 >= 8) {
    y = (long)(x / 1e8);
    ptr += StrPrintF(ptr, "%ld", y);
    x -= (double)y * 1e8;
    lg10 = ((x > 9.) ? (int)log10(x) : 0);
    if (lg10 < 7)
      for (i = 1; i <= (8 - lg10 - 1); i++) *(ptr++) = '0';
  }
  ptr += StrPrintF(ptr, "%ld", (long)x);
	
  *ptr = '\0';
  return 0;
}


static void AddThousandSeparator(char* s)
{
  char buf1[MaxLen__Ans], buf2[MaxLen__Ans];
  char* p0;
  char* p1;
  char* p2;
  int i;

  // skip positive/negative sign. 
  if ((*s < '0') || (*s > '9')) s++;


  *buf1 = '\0';
  *buf2 = '\0';

  p0 = StrChr(s, '.');
  if (p0 == NULL) {
    p0 = s + StrLen(s);
  } else {
    // fraction part 
    p1 = p0 + 1;
    p2 = buf2;
    i = 0;
    while (*p1 != '\0') {
      if ((i > 0) && ((i % 3) == 0) && (ThousSepFracPart != '\0'))
	*(p2++) = ThousSepFracPart;
      *(p2++) = *(p1++);
      i++;
    }
    *p2 = '\0';
  }

  // integer part 
  p1 = p0 - 1;
  p2 = buf1;
  i = 0;
  while (p1 >= s) {
    if ((i > 0) && ((i % 3) == 0) && (ThousSepIntPart != '\0'))
      *(p2++) = ThousSepIntPart;
    *(p2++) = *(p1--);
    i++;
  }
  *p2 = '\0';

  p1 = s;
  p2 = buf1 + StrLen(buf1) - 1;
  while (p2 >= buf1) *(p1++) = *(p2--);
  *(p1++) = '.';
  p2 = buf2;
  while (*p2 != '\0') *(p1++) = *(p2++);
  *p1 = '\0';
}

//=======================================================================================================================================
void formatResult(double cVal, ExprEvalErr_t cErr,                                             //IN 
									AnswerFormat_t cAnsFmt, int cMaxFracDigit, Boolean isShowPrfx, UInt16 widPix, //IN 
									char* strRet){ //Char strRet[MaxLen__Ans];Ƃė̈mۍς݂ł邱       //OUT 
	//MyGlobals_t* gP = getMyGlobals();
	char strBufW[MaxLen__Ans*2];
	UInt16 tStrLen;

	if(cMaxFracDigit < 1) cMaxFracDigit = MaxAnsDig;

	if(cErr == EE_NotEval){
		StrCopy(strBufW, "?");
	}
	else if(cErr != EE_NONE){
		MemHandle mhStr = DmGet1Resource('tSTR', SrID_EE_HEADER);
		if(mhStr == NULL){
			StrPrintF(strBufW, "Err%d", (int)cErr); 
		}else{ 
			StrPrintF(strBufW, "%s%d", (Char*)MemHandleLock(mhStr), (int)cErr); 
			MemHandleUnlock(mhStr); 
			mhStr = NULL;
			switch(cErr){
				case EE_DEEP_NESTING     : mhStr = DmGet1Resource('tSTR', SrID_EE_DEEP_NESTING     ); break;
				case EE_PAREN_NOT_MATCH  : mhStr = DmGet1Resource('tSTR', SrID_EE_PAREN_NOT_MATCH  ); break;
				case EE_NOT_A_NUMBER     : mhStr = DmGet1Resource('tSTR', SrID_EE_NOT_A_NUMBER     ); break;
				case EE_NULL_FOUND       : mhStr = DmGet1Resource('tSTR', SrID_EE_NULL_FOUND       ); break;
				case EE_EXCESS_CHARACTERS: mhStr = DmGet1Resource('tSTR', SrID_EE_EXCESS_CHARACTERS); break;
				case EE_VAR_UNDEFINED    : mhStr = DmGet1Resource('tSTR', SrID_EE_VAR_UNDEFINED    ); break;
			//case EE_DivByZero        : mhStr = DmGet1Resource('tSTR', SrID_EE_DivByZero        ); break;
			//case EE_DeepEvalNest     : mhStr = DmGet1Resource('tSTR', SrID_EE_DeepEvalNest     ); break;
				case EE_RecursiveRef     : mhStr = DmGet1Resource('tSTR', SrID_EE_RecursiveRef     ); break;
				case EE_NullInput        : mhStr = DmGet1Resource('tSTR', SrID_EE_NullInput        ); break;
				case EE_SyntaxErr        : mhStr = DmGet1Resource('tSTR', SrID_EE_SyntaxErr        ); break;
				case EE_DeepRefPage      : mhStr = DmGet1Resource('tSTR', SrID_EE_DeepRefPage      ); break;
				case EE_NoHeapMem        : mhStr = DmGet1Resource('tSTR', SrID_EE_NoHeapMem        ); break;
				default:                   mhStr = NULL; break;
			}
			if(mhStr != NULL){ 
				StrCat(strBufW, (Char*)MemHandleLock(mhStr));
				MemHandleUnlock(mhStr); mhStr = NULL;
			}
		}
	}
	/*
  if(cErr != EE_NONE){
		switch(cErr){
			case EE_NotEval          : StrCopy(strBufW, "?"          ); break;
			case EE_DEEP_NESTING     : StrPrintF(strBufW, "Err%dDeepNest"   , (int)cErr); break;
			case EE_PAREN_NOT_MATCH  : StrPrintF(strBufW, "Err%dParenErr"   , (int)cErr); break;
			case EE_NOT_A_NUMBER     : StrPrintF(strBufW, "Err%dNotNumber"  , (int)cErr); break;
			case EE_NULL_FOUND       : StrPrintF(strBufW, "Err%dNullFound"  , (int)cErr); break;
			case EE_EXCESS_CHARACTERS: StrPrintF(strBufW, "Err%dSyntaxErr"  , (int)cErr); break;
			case EE_VAR_UNDEFINED    : StrPrintF(strBufW, "Err%dNoVar"      , (int)cErr); break;
			case EE_DeepEvalNest     : StrPrintF(strBufW, "Err%dDeepNestVar", (int)cErr); break;
			case EE_RecursiveRef     : StrPrintF(strBufW, "Err%dRecursive"  , (int)cErr); break;
		//default:                   StrPrintF(strBufW, "ErrCode%02d"     , (int)cErr); break;
			default:                   StrPrintF(strBufW, "Err%d"           , (int)cErr); break;
		}
	}//if(cErr != EE_NONE)
	*/
	else if(isnan(cVal)){ //if(cErr == EE_NONE && isnan(cVal))
    StrCopy(strBufW, "[NaN]");
	}
	else if(isinf(cVal) == -1){
    StrCopy(strBufW, "[-INF]");
	}
	else if(isinf(cVal)){
    StrCopy(strBufW, "[INF]");
	}
	else{ //if(cErr == EE_NONE)
			int cFracDigit = cMaxFracDigit;
			int nNumDigit = 0;
			//int nExpDigit = 1;
			double cAbs = fabs(cVal);
			double tVal = cAbs;
			
			{ // 1.23e-45  [e-45] ̕̕(nExpDigit)  [1.23]̕̕(nNumDigit) 𐔂
				int tDigit;
				tVal = cAbs;
				if(     cAbs == 0.){ nNumDigit = 1; }
				else if(cAbs  > 1.){ while(tVal > 1.){ tVal /= 10.; nNumDigit++; }              }
				else if(cAbs  < 1.){ while(tVal < 1.){ tVal *= 10.; nNumDigit++; } } //nExpDigit++; }
				else nNumDigit = 1;
				tDigit = nNumDigit;
				while(tDigit >= 1){ tDigit /= 10; }// nExpDigit++; }
			}
			
			// 0.000000000001 Ƃŕ\͈͂ 0 ȂꍇCIɎw`ɂ
			if((cAnsFmt == AF_f || cAnsFmt == AF_g) && nNumDigit > cMaxFracDigit && cAbs < 1.) 
				cAnsFmt = AF_e;
			
			//{Char cfr[64];StrPrintF(cfr," -%d:%d- \0",(int)cMaxFracDigit,(int)cAnsFmt);WinDrawChars(cfr,StrLen(cfr),70,1);}
			switch(cAnsFmt){
				case AF_g: 
					for(cFracDigit = cMaxFracDigit; cFracDigit > 0; cFracDigit--){
						if(nNumDigit >= cFracDigit){
							//ConvertDblToStr(cVal, strBufW, EXP_FORMAT, cFracDigit-nExpDigit);
							ConvertDblToStr(cVal, strBufW, EXP_FORMAT, cFracDigit);
						}
						else{
							Char strTmpW[MaxLen__Ans * 2];
							//ConvertDblToStr(cVal, strTmpW, EXP_FORMAT, cFracDigit-nExpDigit);
							ConvertDblToStr(cVal, strTmpW, EXP_FORMAT, cFracDigit);
							ConvertDblToStr(cVal, strBufW, FIX_FORMAT, cFracDigit);
							if(StrLen(strTmpW) < StrLen(strBufW))
								//ConvertDblToStr(cVal, strBufW, EXP_FORMAT, cFracDigit-nExpDigit); //StrCopy(strBufW, strTmpW);
								ConvertDblToStr(cVal, strBufW, EXP_FORMAT, cFracDigit); //StrCopy(strBufW, strTmpW);
						}
						if(FldWordWrap(strBufW, widPix) >= StrLen(strBufW)) // \sNZ̃`FbN
							break;                                          // \Ɏ܂ȂΌČJԂ
					}//for(cFracDigit = 12; cFracDigit > 0; cFracDigit--)
					if(FldWordWrap(strBufW, widPix) < StrLen(strBufW))
						goto FORMAT_EXPONENT;
					break; //case AF_g:
					
				default:
				case AF_f: 
					for(cFracDigit = cMaxFracDigit; cFracDigit > 0; cFracDigit--){
						ConvertDblToStr(cVal, strBufW, FIX_FORMAT, cFracDigit); 
						if(cFracDigit == cMaxFracDigit && cAbs > 1. && nNumDigit > FldWordWrap(strBufW, widPix)) goto FORMAT_EXPONENT;
						if(FldWordWrap(strBufW, widPix) >= StrLen(strBufW) && (nNumDigit <= cFracDigit || cAbs >= 1.))
							break;                                         
					}
					if(FldWordWrap(strBufW, widPix) >= StrLen(strBufW) && (nNumDigit <= cFracDigit || cAbs >= 1.))
						break;//case AF_f: //̏ꍇIɎw`ŕ\
					
				case AF_e:
			FORMAT_EXPONENT:
			    // if(cMaxFracDigit-nExpDigit <= 0) cMaxFracDigit = nExpDigit+1;
					// for(cFracDigit = cMaxFracDigit; cFracDigit-nExpDigit > 0; cFracDigit--){
					// 	ConvertDblToStr(cVal, strBufW, EXP_FORMAT, cFracDigit-nExpDigit);
					// 	if(FldWordWrap(strBufW, widPix) >= StrLen(strBufW))
					// 		break;                                         
					// }
			    if(cMaxFracDigit <= 0) cMaxFracDigit = 1;
					for( cFracDigit = cMaxFracDigit ; cFracDigit > 0 ; cFracDigit-- ){
						ConvertDblToStr(cVal, strBufW, EXP_FORMAT, cFracDigit);
						if(FldWordWrap(strBufW, widPix) >= StrLen(strBufW))
							break;      
						//if(StrLen(strBufW) - FldWordWrap(strBufW, widPix) > 0 && cFracDigit > StrLen(strBufW) - FldWordWrap(strBufW, widPix)) 
						//	cFracDigit -= StrLen(strBufW) - FldWordWrap(strBufW, widPix) ;
						//else 
						//	cFracDigit--;
					}
					break;//case AF_e:
					
				case AF_h: 
			FORMAT_16:
					if((unsigned long)cAbs <= 0xFFFF) StrPrintF(strBufW,  "%s0x%x", (cVal<0.) ? "-": "", (unsigned short)cAbs); 
					else                              StrPrintF(strBufW, "%s0x%lx", (cVal<0.) ? "-": "", (unsigned long )cAbs); 
					if(FldWordWrap(strBufW, widPix) < StrLen(strBufW)){
						StrPrintF(strBufW, "%s0x%x", (cVal<0.) ? "-": "", (unsigned short)cAbs); 
					}
					if(FldWordWrap(strBufW, widPix) < StrLen(strBufW))
						goto FORMAT_EXPONENT;
					break;
					
				case AF_b: 
					if((unsigned long)cAbs > 0x00FF) goto FORMAT_16;
					else{
						int cBin;
						unsigned long bMask = 0x1, bMaskMax = 0x0100;
						if((unsigned long)cAbs > 0x000F){	StrPrintF(strBufW, "0b00000000"); bMaskMax = 0x0100; }
						else                            { StrPrintF(strBufW, "0b0000"    ); bMaskMax = 0x0010; }
						for(bMask = 0x1, cBin = StrLen(strBufW)-1; bMask < bMaskMax && cBin >= 2; bMask <<= 1, cBin--){
							if( (unsigned long)cAbs & bMask )
								strBufW[cBin] = '1';
						}
						if(cVal<0.){
							Char strTmpW[MaxLen__Ans+2];
							StrCopy(strTmpW, strBufW);
							StrCopy(strBufW, "-");
							StrCat(strBufW, strTmpW);
						}
					}
					if(FldWordWrap(strBufW, widPix) < StrLen(strBufW))
						goto FORMAT_16;
					break;//case AF_b: 
					
				case AF_Round:
					StrPrintF(strBufW, "%s%ld", ( (isShowPrfx) ? "r." : "" ), (long)round(cVal));
					if(FldWordWrap(strBufW, widPix) < StrLen(strBufW))
						goto FORMAT_EXPONENT;
					break;
					
				case AF_Floor:
					StrPrintF(strBufW, "%s%ld", ( (isShowPrfx) ? "f." : "" ), (long)floor(cVal));
					if(FldWordWrap(strBufW, widPix) < StrLen(strBufW))
						goto FORMAT_EXPONENT;
					break;
					
				case AF_Ceil:
					StrPrintF(strBufW, "%s%ld", ( (isShowPrfx) ? "c." : "" ), (long)ceil(cVal));
					if(FldWordWrap(strBufW, widPix) < StrLen(strBufW))
						goto FORMAT_EXPONENT;
					break;
					
			}//switch(cAnsFmt)

	}//if(cErr == EE_NONE)


	 // \sNZɂ킹
	 tStrLen = FldWordWrap(strBufW, widPix);
	 if(tStrLen >= MaxLen__Ans) tStrLen = MaxLen__Ans-1;
	 //strBufW[tStrLen] = '\0';
	 StrNCopy(strRet, strBufW, tStrLen);
	 strRet[tStrLen] = '\0';
	
}


//EOF
